-- TODO: Handle creation and deletion of avatar nodes when publishing to :config
-- TODO: Somehow make the hosts aware of our "presence"

local host = module:get_host();
if module:get_host_type() ~= "component" then
    error("MIX should be loaded as a component", 0);
end

local st = require("util.stanza");
local jid = require("util.jid");
local uuid = require("util.uuid");
local id = require("util.id");
local datetime = require("util.datetime");
local time = require("util.time");
local dataforms = require("util.dataforms");
local array = require("util.array");
local set = require("util.set");
local lib_pubsub = module:require("pubsub");

local helpers = module:require("mix/helpers");
local namespaces = module:require("mix/namespaces");
local lib_forms = module:require("mix/forms");
local lib_mix = module:require("mix/mix");

local Channel = lib_mix.Channel;
local Participant = lib_mix.Participant;

-- Persistent data
local persistent_channels = module:open_store("mix_channels", "keyval");
local persistent_channel_data = module:open_store("mix_data", "keyval");
local message_archive = module:open_store("mix_log", "archive");

-- Configuration
local default_channel_description = module:get_option("default_description", "A MIX channel for chatting");
local default_channel_name = module:get_option("default_name", "MIX channel");
local restrict_channel_creation = module:get_option("restrict_local_channels", "local");
local service_name = module:get_option("service_name", "Prosody MIX service");

-- MIX configuration
local default_mix_nodes = array { namespaces.info, namespaces.participants, namespaces.messages };

local channels = {};

local function find_channel(channel_jid)
    -- Return the channel object from the channels array for which the
    -- JID matches. If none is found, returns -1, nil
    local _, channel = helpers.find(channels, function(c) return c.jid == channel_jid; end);
    return channel;
end

local function save_channels()
    module:log("debug", "Saving channel list...");
    local channel_list = {};
    for _, channel in pairs(channels) do
        table.insert(channel_list, channel.jid);

        persistent_channel_data:set(channel.jid, channel);
    end

    persistent_channels:set("channels", channel_list);
    module:log("debug", "Saving channel list done.");
end

function Channel:save_state()
    -- Store the channel in the persistent channel store
    module:log("debug", "Saving channel %s...", self.jid);
    persistent_channel_data:set(self.jid, self);
    module:log("debug", "Saving done.", self.jid);
end

function module.load()
    module:log("info", "Loading MIX channels...");

    local channel_list = persistent_channels:get("channels");
    if channel_list then
        for _, channel_data in pairs(channel_list) do
            local channel = Channel:from(persistent_channel_data:get(channel_data));
            table.insert(channels, channel);
            module:log("debug", "MIX channel %s loaded", channel.jid);
        end
    else
        module:log("debug", "No MIX channels found.");
    end
    module:log("info", "Loading MIX channels done.");
end

-- PubSub logic
local function pubsub_not_implemented(stanza, feature)
    local reply = st.error_reply(stanza, "cancel", "feature-not-implemented");
    reply:tag("unsupported", {
        xmlns = namespaces.pubsub_error;
        feature = feature;
    });

    return reply;
end

local function handle_pubsub_iq(event)
	local stanza, origin = event.stanza, event.origin;
    local from = jid.bare(stanza.attr.from);

    local channel = find_channel(stanza.attr.to);
    if not channel then
        module:log("error", "PubSub was used for unknown channel");
        origin:send(st.error_reply(stanza, "cancel", "item-not-found"));
        return;
    end;

    -- Certain actions we do not want the user to perform, so we need to
    -- catch them here.
    local pubsub = stanza:get_child("pubsub", namespaces.pubsub);
    if pubsub then
        local items = pubsub:get_child("items");
        if items then
            if not channel:may_retrieve_items(from, items.attr.node) then
                origin:send(pubsub_not_implemented(stanza, "retrieve-items"));
                return true;
            end
        end

        local publish = pubsub:get_child("publish");
        if publish then
            if not channel:may_publish(from, publish.attr.node) then
                origin:send(pubsub_not_implemented(stanza, "publish"));
                return true;
            end
        end

        local retract = pubsub:get_child("retract");
        if retract then
            if not channel:may_retract(from, retract.attr.node) then
                origin:send(pubsub_not_implemented(stanza, "delete-items"));
                return true;
            end
        end

        -- We generally do not allow deletion, creation or configuration of
        -- nodes. (Un)Subscribing is not allowed as this is managed via
        -- interaction with the MIX host.
        -- NOTE: Checking for <delete> is not needed as no user is ever set as
        --       owner
        if pubsub:get_child("configure") then
            origin:send(pubsub_not_implemented(stanza, "config-node"));
            return true;
        end
        if pubsub:get_child("unsubscibe") or pubsub:get_child("subscribe") then
            origin:send(st.error_reply(stanza, "auth", "forbidden"));
            return true;
        end
    end

    -- TODO: Process publishing to :config

	local service = channel:get_pubsub_service();
	return lib_pubsub.handle_pubsub_iq(event, service);
end
module:hook("iq/bare/"..namespaces.pubsub..":pubsub", handle_pubsub_iq, 1000);

local function can_create_channels(user)
    -- Returns true when the jid is allowed to create MIX channels. False otherwise.
    -- NOTE: Taken from plugins/muc/mod_muc.lua
    local host_suffix = host:gsub("^[^%.]+%.", "");

    if restrict_channel_creation == "local" then
        module:log("debug", "Comparing %s (Sender) to %s (Host)", jid.host(user), host_suffix);

        if jid.host(user) == host_suffix then
            return true;
        else
            return false;
        end
    elseif type(restrict_channel_creation) == "table" then
        if helpers.find_str(restrict_channel_creation, user) ~= -1 then
            -- User was specifically listed
            return true;
        elseif helpers.find_str(restrict_channel_creation, jid.host(user)) then
            -- User's host was allowed
            return true;
        end

        return false;
    end

    -- TODO: Handle also true/"admin" (See mod_muc)
    return true;
end


-- Disco related functionality
module:hook("host-disco-items", function(event)
    module:log("debug", "host-disco-items called");
    local reply = event.reply;
    for _, channel in pairs(channels) do
        -- Adhoc channels are supposed to be invisible
        if not channel.adhoc then
            reply:tag("item", { jid = channel.jid }):up();
        end
    end
end);

local function handle_channel_disco_items(event)
    module:log("debug", "IQ-GET disco#items");

    local origin, stanza = event.origin, event.stanza;
    if stanza:get_child("query", "http://jabber.org/protocol/disco#items").attr.node ~= "mix" then
        origin.send(st.error_reply(stanza, "modify", "bad-request"));
        return true;
    end

    -- TODO: Maybe here we should check if the user has permissions to get infos
    -- about the channel before saying that it doesn't exist to prevent creating
    -- an oracle.
    local channel = find_channel(stanza.attr.to);
    if not channel then
        origin.send(lib_mix.channel_not_found(stanza));
        return true;
    end

    if not channel:is_participant(jid.bare(stanza.attr.from)) then
        origin.send(st.error_reply(stanza, "cancel", "forbidden"));
        return true;
    end

    local reply = st.reply(stanza):tag("query", { xmlns = "http://jabber.org/protocol/disco#items", node = "mix" });
    for _, node in pairs(channel.nodes) do
        reply:tag("item", { jid = channel.jid, node = node }):up();
    end

    origin.send(reply);
    return true;
end
module:hook("iq/bare/http://jabber.org/protocol/disco#items:query", handle_channel_disco_items);

module:hook("iq-get/host/http://jabber.org/protocol/disco#info:query", function(event)
    module:log("debug", "IQ-GET host disco#info");

    local origin, stanza = event.origin, event.stanza;
    local reply = st.reply(stanza)
                    :tag("query", { xmlns = "http://jabber.org/protocol/disco#info" })
                        :tag("identity", { category = "conference", type = "mix", name = service_name }):up()
                        :tag("feature", { var = "http://jabber.org/protocol/disco#info" }):up()
                        :tag("feature", { var = "http://jabber.org/protocol/disco#items" }):up()
                        :tag("feature", { var = namespaces.mix_core }):up();

    if can_create_channels(stanza.attr.from) then
        reply:tag("feature", { var = namespaces.mix_core.."#create-channel" }):up();
    end
    origin.send(reply);
    return true;
end, 1000);

local function handle_channel_disco_info(event)
    module:log("debug", "IQ-GET disco#info");

    local origin, stanza = event.origin, event.stanza;
    local channel = find_channel(stanza.attr.to);
    if not channel then
        origin.send(lib_mix.channel_not_found(stanza));
        return true;
    end
    local reply = st.reply(stanza):tag("query", { xmlns = "http://jabber.org/protocol/disco#info" });
    reply:tag("feature", { var = "http://jabber.org/protocol/disco#info" }):up();
    reply:tag("identity", { category = "conference", name = channel.name, type = "mix" }):up();

    reply:tag("feature", { var = namespaces.mix_core }):up();
    reply:tag("feature", { var = "urn:xmpp:mam:2" }):up();

    origin.send(reply);
    return true;
end
module:hook("iq-get/bare/http://jabber.org/protocol/disco#info:query", handle_channel_disco_info);

module:hook("iq-set/bare/"..namespaces.mam..":query", function(event)
    local stanza, origin = event.stanza, event.origin;
    local channel_jid = stanza.attr.to;
    local channel = find_channel(channel_jid);
    if not channel then
        -- TODO: Is this correct?
        origin.send(lib_mix.channel_not_found(stanza));
        return true;
    end

    -- Check if the user is subscribed to the messages node
    if not channel:is_subscribed(stanza.attr.from, namespaces.messages) then
        origin.send(st.error_reply(stanza, "cancel", "forbidden"));
        return true;
    end

    local query = stanza:get_child("query", namespaces.mam);
    local filter = {};
    local query_id = query.attr.queryid;
    local x = query:get_child("x", "jabber:x:data");
    if x ~= nil then
        -- TODO: Error handling
        local form, err = lib_forms.mam_query:data(x);

        -- Validate
        if (form["start"] and not form["end"]) or (not form["start"] and form["end"]) then
            origin.send(st.error_reply(stanza, "modify", "bad-request", "Invalid timestamp"));
            return true;
        end

        module:log("debug", "Got a MAM query between %s and %s", form["start"], form["end"]);
        filter = {
            start = datetime.parse(form["start"]); ["end"] = datetime.parse(form["end"])
        };
    end
    local data, err = message_archive:find(channel_jid, filter);
    if not data then
        module:log("debug", "MAM error: %s", err);

        if err == "item-not-found" then
            origin.send(st.error_reply(stanza, "modify", "item-not-found"));
        else
            origin.send(st.error_reply(stanza, "cancel", "internal-server-error"));
        end
        return true;
    end
    for message_id, item, when in data do
        local msg = st.stanza("message", { from = channel_jid, to = stanza.attr.from, type = "groupchat" })
                        :tag("result", { xmlns = namespaces.mam, queryid = query_id, id = message_id })
                            :tag("forwarded", { xmlns = "urn:xmpp:forward:0" })
                                :tag("delay", { xmlns = "urn:xmpp:delay", stamp = datetime.datetime(when) }):up();
        msg:add_child(item);
        origin.send(msg);
    end
    return true;
end);

module:hook("iq-get/bare/"..namespaces.mam..":query", function(event)
    if event.stanza.attr.id ~= "form1" then return; end

    module:log("debug", "Got a MAM query for supported fields");

    -- TODO: Use dataforms:...
    local ret = st.reply(event.stanza)
                    :tag("query", { xmlns = namespaces.mam })
                        :tag("x", { xmlns = "jabber:x:data", type = "form"})
                            :tag("field", { type = "hidden", var = "FORM_TYPE" })
                                :tag("value"):text(namespaces.mam):up():up()
                            :tag("field", { type = "jid-single", var = "with" }):up()
                            :tag("field", { type = "text-single", var = "start" }):up()
                            :tag("field", { type = "text-single", var = "end" });
    module:send(ret);
    return true;
end);

module:hook("iq-set/bare/"..namespaces.mix_core..":leave", function(event)
    module:log("debug", "MIX leave received");
    local origin, stanza = event.origin, event.stanza;
    local from = jid.bare(stanza.attr.from);
    local channel = find_channel(stanza.attr.to);
    if not channel then
        origin.send(lib_mix.channel_not_found(stanza));
        return true;
    end

    local participant = channel:find_participant(from);
    if not participant then
        origin.send(st.error_reply(stanza,
                                   "cancel",
                                   "forbidden",
                                   "Not a participant"));
        channel:debug_print();
        module:log("debug", "%s is not a participant in %s", from, channel.jid);
        return true;
    end

    channel:remove_participant(from);

    module:fire_event("mix-channel-leave", { channel = channel, participant = participant });

    origin.send(st.reply(stanza):tag("leave", { xmlns = namespaces.mix_core }));
    return true;
end);

module:hook("iq-set/bare/"..namespaces.mix_core..":join", function(event)
    module:log("debug", "MIX join received");

    local origin, stanza = event.origin, event.stanza;
    local from = jid.bare(stanza.attr.from);
    local channel = find_channel(stanza.attr.to);
    if not channel then
        origin:send(lib_mix.channel_not_found(stanza));
        return true;
    end

    -- Prevent the user from joining multiple times
    local participant = channel:find_participant(from);
    if participant then
        module:send(st.error_reply(stanza, "cancel", "bad-request", "User already joined"));
        return true;
    end

    -- Is the user allowed to join?
    if not channel:may_join(from) then
        origin:send(st.error_reply(stanza, "cancel", "forbidden", "User or host is banned"));
        return true;
    end

    local spid = channel:get_spid(from) or uuid.generate(); -- Stable Participant ID
    local reply = st.reply(stanza)
                    :tag("join", { xmlns = namespaces.mix_core, id = spid });
    local join = stanza:get_child("join", namespaces.mix_core);
    local nick_tag = join:get_child("nick");

    -- TODO: Check if the channel has mandatory nicks

    local nick;
    if not nick_tag then
        nick = jid.node(from);
    else
        nick = nick_tag:get_text();
    end
    module:log("debug", "User joining as nick %s", nick);

    local srv = channel:get_pubsub_service(jid.node(channel.jid));
    local nodes = {};
    local has_subscribed_once = false;
    local first_error = nil;
    local owner_or_admin = channel:is_admin(from) or channel:is_owner(from);
    for subscribe in join:childtags("subscribe") do
        -- May the user subscribe to the node?
        module:log("debug", "Subscribing user to node %s", subscribe.attr.node);
        if channel:may_subscribe(from, subscribe.attr.node, true) then
            local ok, err = srv:add_subscription(subscribe.attr.node, true, from);
            if not ok then
                module:log("debug", "Error during subscription: %s", err);

                -- MIX-CORE says that the first error should be returned when
                -- no of the requested nodes could be subscribed to
                if first_error ~= nil then
                    first_error = err;
                end
            else
                table.insert(nodes, subscribe.attr.node);
                reply:tag("subscribe", { node = subscribe.attr.node }):up();
                has_subscribed_once = true;
            end

            -- Set the correct affiliation
            channel:set_affiliation(subscribe.attr.node, from, "member");
        else
            module:log("debug", "Error during subscription: may_subscribe returned false");
            if first_error ~= nil then
                first_error = "Channel does not allow subscribing";
            end
        end
    end

    if not has_subscribed_once then
        -- TODO: This does not work
        origin:send(st.error_reply(stanza, "cancel", first_error));
        return true;
    end

    -- TODO: Participant configuration

    local participant = Participant:new(jid.bare(from), nick, {});
    channel.subscriptions[from] = nodes;
    table.insert(channel.participants, participant)
    channel:set_spid(jid.bare(stanza.attr.from), spid);
    channel:publish_participant(spid, participant);
    channel:save_state();

    module:fire_event("mix-channel-join", { channel = channel, participant = participant });

    -- We do not reuse nick_tag as it might be nil
    reply:tag("nick"):text(nick):up();
    reply:add_child(anon.form:form(config, "result"));
    origin.send(reply);
    return true
end);

module:hook("iq-set/bare/"..namespaces.mix_core..":setnick", function(event)
    module:log("debug", "MIX setnick received");
    local origin, stanza = event.origin, event.stanza;
    local from = jid.bare(stanza.attr.from);
    local channel = find_channel(stanza.attr.to);
    if not channel then
        origin.send(lib_mix.channel_not_found(stanza));
        return true;
    end

    local participant = channel:find_participant(from);
    if not participant then
        channel:debug_print();
        module:log("debug", "%s is not a participant in %s", from, channel.jid);
        return true;
    end

    -- NOTE: Prosody should guarantee us that the setnick stanza exists
    local setnick = stanza:get_child("setnick", namespaces.mix_core);
    local nick = setnick:get_child("nick");
    if nick == nil then
        origin.send(st.error_reply(stanza, "cancel", "bad-request", "Missing <nick>"));
        return true;
    end

    -- Change the nick
    participant.nick = nick:get_text();
    -- Inform all other members
    channel:publish_participant(channel:get_spid(participant.jid), participant);

    module:fire_event("mix-change-nick", { channel = channel, participant = participant });

    origin.send(st.reply(stanza)
                    :tag("setnick", { xmlns = namespaces.mix_core })
                        :tag("nick"):text(nick:get_text()));
    channel:save_state();
    return true;
end);

local function create_channel(node, creator, adhoc)
    -- TODO: Now all properties from the admin dataform are covered
    local channel = Channel:new(string.format("%s@%s", node, host), -- Channel JID
                                default_channel_name,
                                default_channel_description,
                                {},             -- Participants
                                {},             -- Administrators
                                { creator },    -- Owners
                                {},             -- Subscriptions
                                {},             -- SPID mapping
                                { creator },    -- Contacts
                                adhoc,          -- Is channel an AdHoc channel
                                {},             -- Allowed
                                {},             -- Banned
                                lib_mix.default_channel_configuration, -- Channel config
                                {});            -- Present nodes

    -- Create the PubSub nodes
    local srv = channel:get_pubsub_service();
    for _, psnode in ipairs(default_mix_nodes) do
        srv:create(psnode, true, {
            -- NOTE: Our custom PubSub service is persistent only, so we don't
            --       need to explicitly set it
            ["access_model"] = lib_mix.get_node_access_model(psnode, adhoc);
            ["max_items"] = lib_mix.get_node_max_items(psnode);
        });
        channel:set_affiliation(psnode, creator, "creator");
        table.insert(channel.nodes, psnode);
    end

    channel:publish_info(srv);
    table.insert(channels, channel);
end

module:hook("iq-set/host/"..namespaces.mix_core..":create", function(event)
    module:log("debug", "MIX create received");
    local origin, stanza = event.origin, event.stanza;
    local from = jid.bare(stanza.attr.from);

    -- Check permissions
    if not can_create_channels(from) then
        origin.send(st.error_reply(stanza, "cancel", "forbidden", "Not authorized to create channels"));
        return true;
    end

    local create = stanza:get_child("create", namespaces.mix_core);
    local node;
    if create.attr.channel ~= nil then
        -- Create non-adhoc channel
        module:log("debug", "Attempting to create channel %s", create.attr.channel);
        node = create.attr.channel;
        local channel = find_channel(create.attr.channel.."@"..stanza.attr.to);
        if channel then
            origin.send(st.error_reply(stanza,
                                       "cancel",
                                       "conflict",
                                       "Channel already exists"));
            return true;
        end

        create_channel(create.attr.channel, from, false);
    else
        -- Create adhoc channel
        while (true) do
            node = id.short();
            local ch = find_channel(string.format("%s@%s", node, host));
            if not ch then
                break;
            end
        end

        create_channel(node, from, true);
    end
    module:log("debug", "Channel %s created with %s as owner", node, from);
    -- TODO: Add an event

    origin.send(st.reply(stanza)
                :tag("create", { xmlns = namespaces.mix_core, channel = node }));
    save_channels();
    return true;
end);

module:hook("iq-set/host/"..namespaces.mix_core..":destroy", function(event)
    module:log("debug", "MIX destroy received");
    local origin, stanza = event.origin, event.stanza;
    local from = jid.bare(stanza.attr.from);

    local destroy = stanza:get_child("destroy", namespaces.mix_core);
    local node = destroy.attr.channel;
    local node_jid = string.format("%s@%s", node, host);
    local channel = find_channel(node_jid);
    if not channel then
        origin.send(lib_mix.channel_not_found(stanza));
        return true;
    end

    -- TODO(MIX-ADMIN): Check if the user is the owner of the channel
    -- Until then, we just check if the user is in the contacts
    if helpers.find_str(channel.contacts, from) == -1 then
        origin.send(st.error_reply(stanza, "cancel", "forbidden"));
        return true;
    end

    if module:fire_event("mix-destroy-channel", { channel = channel }) then
        return true;
    end

    -- Remove all registered nodes
    local srv = channel:get_pubsub_service();
    for _, psnode in pairs(channel.nodes) do
        srv:delete(psnode, true);
    end
    channels = array.filter(channels, function (c) return c.jid ~= node_jid end); 

    module:fire_event("mix-channel-destroyed", { channel = channel });
    module:log("debug", "Channel %s destroyed", node);

    origin.send(st.reply(stanza));
    save_channels();
    return true;
end);

module:hook("message/bare", function(event)
    module:log("debug", "MIX message received");
    local stanza, origin = event.stanza, event.origin;
    if stanza.attr.type ~= "groupchat" then
        origin.send(st.error_reply(stanza, "modify", "bad-request", "Non-groupchat message"));
        return true;
    end

    local from = jid.bare(stanza.attr.from);
    local channel = find_channel(stanza.attr.to);
    if not channel then
        origin.send(lib_mix.channel_not_found(stanza));
        return true;
    end

    local participant = channel:find_participant(from);
    if not participant then
        origin.send(st.error_reply(stanza, "cancel", "forbidden", "Not a participant"));
        return true;
    end

    -- Handles sending the message accordingly, firing an event and
    -- even doing nothing if an event handler for "mix-broadcast-message"
    -- returns true.
    channel:broadcast_message(stanza, participant, message_archive);
    return true;
end);
